// bash-syspose.c - interpose so that bash shell prints $PS4 upon syscall
// Copyright 2011 John Reiser, BitWagon Software LLC.  All rights reserved.
// Licensed under GNU General Public License, version 3 (GPLv3).
//
// Requires: change indirection_level_string() in bash/print_cmd.c
//    to preserve 'x' flag around decode_prompt_string():
/*
--- bash-4.2/print_cmd.c	2011-09-08 13:27:53.877244584 -0700
+++ new/print_cmd.c	2011-09-08 13:29:34.088991764 -0700
@@ -426,9 +426,9 @@
   if (ps4 == 0 || *ps4 == '\0')
     return (indirection_string);
 
-  change_flag ('x', FLAG_OFF);
+  int const old_x = change_flag ('x', FLAG_OFF);
   ps4 = decode_prompt_string (ps4);
-  change_flag ('x', FLAG_ON);
+  change_flag ('x', (old_x ? FLAG_ON : FLAG_OFF));
 
   if (ps4 == 0 || *ps4 == '\0')
     return (indirection_string);
*/
// Build:  gcc -shared -o bash-syspose.so -g -O -fPIC bash-syspose.c -ldl
// Run:  LD_PRELOAD=./bash-syspose.so  bash  ...  # "unset LD_PRELOAD" ASAP!
// EXAMPLE: (insert near beginning of shell script)
//   unset LD_PRELOAD
//   export PS4=' ${FUNCNAME[0]}@$LINENO < ${BASH_SOURCE[1]##*/}:${BASH_LINENO[0]} :'
//

// Change as desired: conditional compilation of each intercept routine.
#define TRACE_access   0
#define TRACE_chdir    1
#define TRACE_close    0
#define TRACE_dup      0
#define TRACE_dup2     0
#define TRACE_execve   1
#define TRACE_fork     1
#define TRACE_fstat    0
#define TRACE_lstat    0
#define TRACE_open     1
#define TRACE_pipe     0
#define TRACE_pipe2    0
#define TRACE_read     0
#define TRACE_readlink 1
#define TRACE_stat     0
#define TRACE_unlink   1

// Unimplemented:
//	alarm
//	connect, socket
//	faccessat
//	fcntl
//	getcwd
//	getpid, getppid, getpgrp
//	getuid, getgid, geteuid, getegid
//	getrlimit, getrusage
//	gettimeofday
//	ioctl
//	kill
//	nanosleep
//	read, write, lseek
//	select
//	setitimer
//	setuid, setgid, setpgid
//	sigprocmask
//	time
//	waitpid (wait4)


#define _GNU_SOURCE 1  /* need RTLD_NEXT from dlfcn.h */
#include <dlfcn.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

// weak: Test to prevent crash if no such function, such as when
// LD_PRELOAD remains set and a child process of bash invokes a
// random executable that calls a TRACEd syscall.
extern char const *indirection_level_string(void) __attribute__((weak));

static unsigned umin(unsigned a, unsigned b)
{
	return (a <= b) ? a : b;
}

static int recur;  // recursion level counter
static struct timeval epoch;  // earliest gettimeofday()

static void
init_epoch()
{
	if (epoch.tv_sec==0)
		gettimeofday(&epoch, 0);
}


static int
get_ils(char *const ilsbuf, int const buflen)
{
	char const *p = (indirection_level_string  // implemented by bash
		? indirection_level_string()  // evaluate PS4
		: "" );
	int const lenp = umin(-1+ buflen, strlen(p));
	memcpy(ilsbuf, p, lenp);
	ilsbuf[lenp] = '\0';
	return lenp;
}

struct Err_dict {
	char const name[8];
};

#define N_ERR_DICT 33
#define xx(sym, num, def) [num] = {#sym}
static struct Err_dict const err_dict[N_ERR_DICT] = {
	xx(ENONE, 0, "Success"),
	xx(EPERM,  1, "Operation not permitted"), 
	xx(ENOENT,  2, "No such file or directory"), 
	xx(ESRCH,  3, "No such process"), 
	xx(EINTR,  4, "Interrupted system call"), 
	xx(EIO,  5, "I/O error"), 
	xx(ENXIO,  6, "No such device or address"), 
	xx(E2BIG,  7, "Argument list too long"), 
	xx(ENOEXEC,  8, "Exec format error"), 
	xx(EBADF,  9, "Bad file number"), 
	xx(ECHILD, 10, "No child processes"), 
	xx(EAGAIN, 11, "Try again"), 
	xx(ENOMEM, 12, "Out of memory"), 
	xx(EACCES, 13, "Permission denied"), 
	xx(EFAULT, 14, "Bad address"), 
	xx(ENOTBLK, 15, "Block device required"), 
	xx(EBUSY, 16, "Device or resource busy"), 
	xx(EEXIST, 17, "File exists"), 
	xx(EXDEV, 18, "Cross-device link"), 
	xx(ENODEV, 19, "No such device"), 
	xx(ENOTDIR, 20, "Not a directory"), 
	xx(EISDIR, 21, "Is a directory"), 
	xx(EINVAL, 22, "Invalid argument"), 
	xx(ENFILE, 23, "File table overflow"), 
	xx(EMFILE, 24, "Too many open files"), 
	xx(ENOTTY, 25, "Not a typewriter"), 
	xx(ETXTBSY, 26, "Text file busy"), 
	xx(EFBIG, 27, "File too large"), 
	xx(ENOSPC, 28, "No space left on device"), 
	xx(ESPIPE, 29, "Illegal seek"), 
	xx(EROFS, 30, "Read-only file system"), 
	xx(EMLINK, 31, "Too many links"), 
	xx(EPIPE, 32, "Broken pipe"), 
};

static void
show_r(
	unsigned long const retval,
	struct timeval now,
	char const *const ils,
	char const *const sysname,
	char const *const fmt,
	...
)
{
	unsigned volatile const e = errno;

	now.tv_sec -= epoch.tv_sec;
	if (0 > (now.tv_usec -= epoch.tv_usec)) {
		now.tv_sec--;
		now.tv_usec += 1000000;
	}

	char buf[400];
	unsigned len = snprintf(buf, sizeof(buf), "%d.%06d %s %d:%s\n",
		now.tv_sec, now.tv_usec, (ils?:""), getpid(), sysname);
	len = umin(len, sizeof(buf));
	if (fmt) { // extra info
		if ('\n'==buf[-1+ len])
			--len;  // lop trailing '\n'
		va_list ap;
		va_start(ap, fmt);
		len += vsnprintf(&buf[len], sizeof(buf) - len, fmt, ap);
		len = umin(len, sizeof(buf));
		va_end(ap);
	}
	if ('\n'==buf[-1+ len])
		--len;  // lop trailing '\n'
	len += snprintf(&buf[len], sizeof(buf) - len, " = %d\n", retval);
	len = umin(len, sizeof(buf));
	if (~0xffful < retval) {
		if ('\n'==buf[-1+ len])
			--len;  // lop trailing '\n'
		if (e < N_ERR_DICT) {
			len += snprintf(&buf[len], sizeof(buf) - len, " %s (%s)\n",
				err_dict[e].name, sys_errlist[e]);
		}
		else {
			len += snprintf(&buf[len], sizeof(buf) - len, " errno=%d\n", e);
		}
		len = umin(len, sizeof(buf));
	}
	buf[-1+ len] = '\n';
	write(2, buf, len);
}

// This macro saves keystrokes, and provides uniformity.
#define part1(sym, args) \
	typedef int (*sym##_t)args; \
	static sym##_t real_##sym; if (!real_##sym) { \
		real_##sym = (sym##_t)dlsym(RTLD_NEXT, #sym); \
		init_epoch(); \
	} \
	struct timeval now; char ilsbuf[200]; \
	if (!recur++) { \
		gettimeofday(&now, 0); \
		get_ils(ilsbuf, sizeof(ilsbuf)); \
	}

#if TRACE_access  /*{*/
int
access(char const *pathname, int mode)
{
	part1(access,(char const *, int));
	int const retval = (*real_access)(pathname, mode);
	if (!--recur)
		show_r(retval, now, ilsbuf, "access", "(\"%s\", %#x)", pathname, mode);
	return retval;
}
#endif  /*}*/

#if TRACE_chdir  /*{*/
int
chdir(char const *const path)
{
	part1(chdir, (char const *));
	int const retval = (*real_chdir)(path);
	if (!--recur)
		show_r(retval, now, ilsbuf, "chdir", "(\"%s\")", path);
	return retval;
}
#endif  /*}*/

#if TRACE_close  /*{*/
int
close(int const oldfd)
{
	part1(close, (int));
	int const retval = (*real_close)(oldfd);
	if (!--recur)
		show_r(retval, now, ilsbuf, "close", "(%d)", oldfd);
	return retval;
}
#endif  /*}*/

#if TRACE_dup  /*{*/
int
dup(int const oldfd)
{
	part1(dup, (int));
	int const retval = (*real_dup)(oldfd);
	if (!--recur)
		show_r(retval, now, ilsbuf, "dup", "(%d)", oldfd);
	return retval;
}
#endif  /*}*/

#if TRACE_dup2  /*{*/
int
dup2(int const oldfd, int const newfd)
{
	part1(dup2, (int, int));
	int const retval = (*real_dup2)(oldfd, newfd);
	if (!--recur)
		show_r(retval, now, ilsbuf, "dup2", "(%d, %d)", oldfd, newfd);
	return retval;
}
#endif  /*}*/

#if TRACE_execve  /*{*/
static char *
argstr(char *const buf, unsigned const buflen, char const *const *argv)
{
	unsigned len = 0;
	buf[len++] = '[';
	for (; ((4+ len) < buflen) && *argv; ++argv) {  // 4: punctuation
		len += snprintf(&buf[len], buflen - len, "\"%s\", ", *argv);
	}
	len = umin(len, buflen);
	if (2 <= len) len -= 2;  // discard ", " separator
	len += snprintf(&buf[len], buflen - len, "%c", ']');  // always fits!
	return buf;
}

static int
envplen(char const *const *envp)
{
	int len = 0;
	for (; *envp; ++envp)
		++len;
	return len;
}

int
execve(char const *const filename, char const *const argv[], char const *const envp[])
{
	typedef int (*execve_t)(char const *, char const *const [], char const *const []);
	static execve_t real_execve; if (!real_execve) {
		real_execve = (execve_t)dlsym(RTLD_NEXT, "execve");
		init_epoch();
	}
	struct timeval now; char ilsbuf[200];
	if (!recur++) {
		gettimeofday(&now, 0);
		get_ils(ilsbuf, sizeof(ilsbuf));
		// execve() does not return on success, so trace it in advance.
		char argbuf[200]; argstr(argbuf, sizeof(argbuf), argv);
		show_r(0, now, ilsbuf, "execve", "(\"%s\", %s, %d vars)\n",
			filename, argbuf, envplen(envp));
	}
	int const retval = (*real_execve)(filename, argv, envp);
	if (!--recur)
		show_r(retval, now, "", "FAILED execve", "(\"%s\")\n", filename);
}
#endif  /*}*/

#if TRACE_fork  /*{*/
int
fork(void)
{
	static int (*real_fork)(void); if (!real_fork) {
		real_fork = (int (*)(void)) dlsym(RTLD_NEXT, "fork");
		init_epoch();
	}
	struct timeval now; char ilsbuf[200];
	if (!recur++) {
		gettimeofday(&now, 0);
		get_ils(ilsbuf, sizeof(ilsbuf));
	}
	int const retval = (*real_fork)();
	if (!--recur) {
		if (retval) {  // parent (or failure)
			show_r(retval, now, ilsbuf, "fork()", 0);
		}
	}
	// The child might want to remove us (bash-syspose) from LD_PRELOAD.
	// But it is ugly, so rely on 'weak' for now.

	return retval;
}
#endif  /*}*/

#if TRACE_open  /*{*/
int
open(const char *const pathname, int const flags, mode_t const mode)
{
	part1(open, (char const *, int, mode_t));
	int const retval = (*real_open)(pathname, flags, mode);
	if (!--recur)
		show_r(retval, now, ilsbuf, "open", "(\"%s\", %#x)", pathname, flags);
	return retval;
}
#endif  /*}*/

#if TRACE_pipe  /*{*/
int
pipe(int fd[2])
{
	part1(pipe, (int [2]));
	int const retval = (*real_pipe)(fd);
	if (!--recur)
		show_r(retval, now, ilsbuf, "pipe", "([%d, %d])", fd[0], fd[1]);
	return retval;
}
#endif  /*}*/

#if TRACE_pipe2  /*{*/
int
pipe2(int fd[2], int flags)
{
	part1(pipe2, (int [2], int));
	int const retval = (*real_pipe2)(fd, flags);
	if (!--recur)
		show_r(retval, now, ilsbuf, "pipe2", "([%d, %d], %d)",
			fd[0], fd[1], flags);
	return retval;
}
#endif  /*}*/

#if TRACE_read  /*{*/
int
read(int const fd, void *const buf, size_t const count)
{
	part1(read, (int, void *, size_t));
	int const retval = (*real_read)(fd, buf, count);
	if (!--recur)
		show_r(retval, now, ilsbuf, "read", "(%d, %p, %u)",
			fd, buf, count);
	return retval;
}
#endif  /*}*/

#if TRACE_readlink  /*{*/
int
readlink(char const *const path, char *const buf, size_t const bufsiz)
{
	part1(readlink, (char const *, char *, size_t));
	int const retval = (*real_readlink)(path, buf, bufsiz);
	if (!--recur)
		show_r(retval, now, ilsbuf, "readlink", "(\"%s\", \"%s\", %d)",
			path, buf, bufsiz);
	return retval;
}
#endif  /*}*/

#if TRACE_unlink  /*{*/
int
unlink(char const *const path)
{
	part1(unlink, (char const *));
	int const retval = (*real_unlink)(path);
	if (!--recur)
		show_r(retval, now, ilsbuf, "unlink", "(\"%s\")", path);
	return retval;
}
#endif  /*}*/

// FIXME: Beware kernel struct stat versus user struct stat
// (the very problem that glibc handles by using the "__x" symbols.)

#if TRACE_stat  /*{*/
int
__xstat(int const ver, char const *const pathname, struct stat *const statbuf)
{
	part1(__xstat, (int, char const *, struct stat *));
	int const retval = (*real___xstat)(ver, pathname, statbuf);
	if (!--recur)
		show_r(retval, now, ilsbuf, "stat", "(\"%s\", %p)", pathname, statbuf);
	return retval;
}
#endif  /*}*/

#if TRACE_lstat  /*{*/
int
__lxstat(int const ver, char const *const pathname, struct stat *const statbuf)
{
	part1(__lxstat, (int, char const *, struct stat *));
	int const retval = (*real___lxstat)(ver, pathname, statbuf);
	if (!--recur)
		show_r(retval, now, ilsbuf, "lstat", "(\"%s\", %p)", pathname, statbuf);
	return retval;
}
#endif  /*}*/

#if TRACE_fstat  /*{*/
int
__fxstat(int const ver, int const fd, struct stat *const statbuf)
{
	part1(__fxstat, (int, int, struct stat *));
	int const retval = (*real_fxstat)(ver, fd, statbuf);
	if (!--recur)
		show_r(retval, now, ilsbuf, "fstat", "(%d, %p)", fd, statbuf);
	return retval;
}
#endif  /*}*/

// end-of-file bash-syspose.c
